/**********************************************************************************
 * Copyright (c) 2019 Process Systems Engineering (AVT.SVT), RWTH Aachen University
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 **********************************************************************************/

#include "bab.h"
#include "MAiNGOException.h"
#include "getTime.h"
#include "lbp.h"
#include "mpiUtilities.h"
#include "ubp.h"
#include "utilities.h"

#include <limits>

using namespace maingo;
using namespace bab;


////////////////////////////////////////////////////////////////////////////////////////
// constructor for the branch and bound object
BranchAndBound::BranchAndBound(const std::vector<babBase::OptimizationVariable> &variables, std::shared_ptr<lbp::LowerBoundingSolver> LBSIn, std::shared_ptr<ubp::UpperBoundingSolver> UBSIn,
                               std::shared_ptr<Settings> settingsIn, std::shared_ptr<Logger> loggerIn, const unsigned nvarWOauxIn):
    _originalVariables(variables),
    _LBS(LBSIn), _UBS(UBSIn), _nvar(variables.size()), _maingoSettings(settingsIn), _logger(loggerIn), _nvarWOaux(nvarWOauxIn)
{

#ifdef HAVE_MAiNGO_MPI
    // Initialize common MPI variables
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &_nProcs);
    _bcastTag = BCAST_NOTHING_PENDING;

    // Initialize manager related variables
    _checkForNodeWithIncumbent = false;
    _confirmedTermination      = false;
    _workCount                 = 0;

    // Initialize worker related variables
    _pendingIncumbentUpdate = false;
    _moreScalingActivated   = false;

    MAiNGO_IF_BAB_MANAGER
        _informedWorkerAboutIncumbent = std::vector<bool>(_nProcs - 1, false);
        _nodesGivenToWorkers          = std::vector<std::pair<bool, double>>(_nProcs - 1, std::make_pair(false, _maingoSettings->infinity));
        MAiNGO_ELSE
            _brancher = nullptr;
        MAiNGO_END_IF

#endif

        MAiNGO_IF_BAB_MANAGER
            //Create brancher
            _brancher = std::unique_ptr<babBase::Brancher>(new babBase::Brancher(variables));
            // Store settings:
            _brancher->set_branching_dimension_selection_strategy(_maingoSettings->BAB_branchVariable);
            _brancher->set_node_selection_strategy(_maingoSettings->BAB_nodeSelection);
            _brancher->decrease_pruning_score_threshold_to(_maingoSettings->infinity);
            _brancher->enable_pruning_with_rel_and_abs_tolerance(_maingoSettings->epsilonR, _maingoSettings->epsilonA);
        MAiNGO_END_IF

        // Initialize internals
        _foundFeas       = false;
        _firstFound      = 0;
        _incumbentNodeId = 0;
        _lbd             = -_maingoSettings->infinity;
        _lbdOld          = -_maingoSettings->infinity;
        _bestLbdFathomed = _maingoSettings->infinity;
        _ubd             = _maingoSettings->infinity;

        // Set counters to 0
        _nNodesTotal          = 0;
        _nNodesLeft           = 0;
        _nNodesDeleted        = 0;
        _nNodesFathomed       = 0;
        _nNodesMaxInMemory    = 0;
        _lbdcnt               = 0;
        _ubdcnt               = 0;
        _timePassed           = 0.0;
        _timePreprocess       = 0.0;
        _daysPassed           = 0;
        _iterations           = 0;
        _linesprinted         = 0;
        _printNewIncumbent    = false;
        _writeToLogEverySec   = 1;                                      // First write to log is after Settings->writeToLogSec CPU seconds
        _iterationsgap        = 20 * _maingoSettings->BAB_printFreq;    // Print after every 20 outputs
        _lbdNotChanged        = 0;
        _moreScalingActivated = false;

        // Save original bounds
        _lowerVarBoundsOrig.resize(_originalVariables.size());
        _upperVarBoundsOrig.resize(_originalVariables.size());
        for (unsigned int i = 0; i < _originalVariables.size(); i++) {
            _lowerVarBoundsOrig[i] = _originalVariables[i].get_lower_bound();
            _upperVarBoundsOrig[i] = _originalVariables[i].get_upper_bound();
        }
}


////////////////////////////////////////////////////////////////////////////////////////
// main branch and bound algorithm
babBase::enums::BAB_RETCODE
BranchAndBound::solve(babBase::BabNode &rootNodeIn, double &solutionValue, std::vector<double> &solutionPoint, const double preprocessTime, double &timePassed)
{
    try {
        // ---------------------------------------------------------------------------------
        // 0: Process Inputs
        // ---------------------------------------------------------------------------------

        // Auxiliary variables for timing (have to be declared outside of MAiNGO_IF_BAB_MANAGER scope to be available in whole loop)
        double startTime, currentTime, lastTime;
        double currentLBD;    // This is declared here to be available for the Manager and Workers in the parallel version

        MAiNGO_IF_BAB_MANAGER
            // Store initial feasible point if given
            if (!solutionPoint.empty()) {
                _update_incumbent_and_fathom(solutionValue, solutionPoint, 0);
                _incumbentNodeId = 1;    // To make sure we track correctly that the incumbent is now in the root node (which will have ID 1, although the solution was found at node 0 (i.e., pre-processing))
                // Since _update_incumbent_and_fathom triggers an asterisk in the output whenever a new incumbent was found,
                // we need to suppress this here (otherwise it would look like this incumbent was found during B&B in the root node, although it comes from pre-processing)
                _printNewIncumbent = false;
            }

            // Hand root node to _brancher
            // Change id from 0 to 1 since we start B&B
            rootNodeIn = babBase::BabNode(rootNodeIn.get_pruning_score(), rootNodeIn.get_lower_bounds(), rootNodeIn.get_upper_bounds(), 1, 0, rootNodeIn.holds_incumbent());
            _brancher->insert_root_node(rootNodeIn);
            _nNodesLeft = _brancher->get_nodes_in_tree();

            // Initialize stuff for timing and logging
            _timePreprocess = preprocessTime;
            _timePassed     = preprocessTime;
            startTime       = get_cpu_time();
            lastTime        = -1e10;    // This is needed since the time is reset after 24h --> if we detect this, we can store it as days passed
#ifdef HAVE_MAiNGO_MPI
            MAiNGO_ELSE    // Worker
                startTime = get_cpu_time();
                // Opening for broadcasts from master
                MPI_Ibcast(&_bcastTag, 1, MPI_INT, 0, MPI_COMM_WORLD, &_bcastReq);
#endif
            MAiNGO_END_IF

            // ------------------------- End 0: Process Inputs ---------------------------------


            // ---------------------------------------------------------------------------------
            // 1-10: Main Branch-and-Bound loop
            // ---------------------------------------------------------------------------------
            MAiNGO_IF_BAB_MANAGER
                _logger->print_message("\n  Entering branch-and-bound loop:\n", _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
            MAiNGO_END_IF

            // Check termination criteria: when terminating regularly, no nodes should be left; otherwise, termination occurs when hitting CPU time or node limit
            _TERMINATION_TYPE termination = _check_termination();    // _workers == 0
            while (termination != _TERMINATED) {

                babBase::BabNode currentNode;

#ifdef HAVE_MAiNGO_MPI
                MPI_Status status;
                MAiNGO_IF_BAB_MANAGER
                    // -----------------------------------
                    // Check incoming messages from workers
                    // -----------------------------------
                    // Don't accept node requests if babTree is empty or terminated
                    if (_nNodesLeft == 0 || termination == _TERMINATED_WORKERS_ACTIVE) {
                        const int tagCount = 5;
                        // Allowed are exceptions and solved nodes
                        COMMUNICATION_TAG allowedTags[tagCount] = {TAG_EXCEPTION, TAG_FOUND_INCUMBENT, TAG_SOLVED_NODE_STATUS_NORMAL,
                                                                   TAG_SOLVED_NODE_STATUS_CONVERGED, TAG_SOLVED_NODE_STATUS_INFEAS};
                        int flag                                = 0;
                        // Probe for messages with allowed tags
                        while (!flag) {
                            for (int i = 0; i < tagCount; i++) {
                                MPI_Iprobe(MPI_ANY_SOURCE, allowedTags[i], MPI_COMM_WORLD, &flag, &status);
                                if (flag)
                                    break;
                            }
                        }
                    }
                    else {
                        // Otherwise just check for any incoming tags
                        MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
                    }

                    if (status.MPI_TAG == TAG_EXCEPTION) {
                        // -----------------------------------
                        // Worker ran into an exception
                        // -----------------------------------
                        MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                        std::string str = "  Worker ran into exception!";
                        throw MAiNGOMpiException(str, MAiNGOMpiException::ORIGIN_OTHER);
                    }
                    else if (status.MPI_TAG == TAG_FOUND_INCUMBENT) {
                        // -----------------------------------
                        // Worker has found new incumbent
                        // -----------------------------------
                        std::vector<double> incumbentBuf(_nvarWOaux + 2);
                        MPI_Recv(incumbentBuf.data(), _nvarWOaux + 2, MPI_DOUBLE, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                        // Store incumbent
                        std::vector<double> newIncumbent(incumbentBuf.begin(), incumbentBuf.begin() + _nvarWOaux);
                        double newUBD        = incumbentBuf[_nvarWOaux];
                        unsigned incumbentID = incumbentBuf[_nvarWOaux + 1];

                        _update_incumbent_and_fathom(newUBD, newIncumbent, incumbentID);

                        continue;
                    }
                    else if (status.MPI_TAG == TAG_NODE_REQUEST) {

                        // -----------------------------------
                        // Worker requests a new node
                        // -----------------------------------
                        MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                        _workCount++;
#endif

                        // -----------------------------------
                        // 1. Node selection: Determine which node to treat next
                        // -----------------------------------
                        currentNode = _brancher->get_next_node();
                        if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
                            _print_one_node(currentNode.get_pruning_score(), currentNode);
                        }

#ifdef HAVE_MAiNGO_MPI

                        // -----------------------------------
                        // Send node to worker
                        // -----------------------------------
                        _send_new_problem(currentNode, status.MPI_SOURCE);
                        _nNodesLeft                                        = _brancher->get_nodes_in_tree();
                        _nodesGivenToWorkers[status.MPI_SOURCE - 1].first  = true;
                        _nodesGivenToWorkers[status.MPI_SOURCE - 1].second = currentNode.get_pruning_score();

                        continue;
                    }
                MAiNGO_END_IF    // End of MAiNGO_IF_BAB_MANAGER
#endif

#ifdef HAVE_MAiNGO_MPI
                MAiNGO_IF_BAB_WORKER
                    // -----------------------------------
                    // Receive new node from Manager
                    // -----------------------------------

                    // Send node request to manager
                    MPI_Request node_request;
                    MPI_Issend(NULL, 0, MPI_INT, 0, TAG_NODE_REQUEST, MPI_COMM_WORLD, &node_request);

                    bool terminate;
                    _sync_with_master(node_request, terminate);
                    if (terminate)
                        break;

                    // Receive new problem and update _incumbent if necessary
                    _recv_new_problem(currentNode);
#endif
                    //-------------------------------------
                    // Process Node:
                    // 2. Node pre-processing:
                    //      a) Constraint propagation
                    //      b) Optimization-based bound tightening (OBBT)
                    // 3. Lower bounding: solve lower bounding problem (LBP) to derive a better lower bound
                    // 4. Upper bounding: try to find a good feasible point within the current node
                    // 5. Node post-processing:
                    //      a) Duality-based bounds tightening (DBBT)
                    //      b) Probing
                    //-------------------------------------
                    bool nodeProvenInfeasible = false;
                    bool nodeConverged        = false;
                    unsigned nAddedLBSolves   = 0;
                    unsigned nAddedUBSolves   = 0;
                    currentLBD                = currentNode.get_pruning_score();
                    std::vector<double> lbpSolutionPoint;
                    bool foundNewFeasiblePoint = false;
                    double ubpObjectiveValue   = _maingoSettings->infinity;
                    std::vector<double> ubpSolutionPoint;
                    std::tie(nodeProvenInfeasible, nodeConverged, nAddedLBSolves, nAddedUBSolves, currentLBD, lbpSolutionPoint, foundNewFeasiblePoint, ubpObjectiveValue, ubpSolutionPoint) = _process_node(currentNode);


#ifdef HAVE_MAiNGO_MPI
                    // -----------------------------------
                    // Send solved node to Manager
                    // -----------------------------------

                    // Wait until manager accepted pending incumbent updates
                    if (_pendingIncumbentUpdate) {
                        _sync_with_master(_incumbentReq, terminate);
                        _pendingIncumbentUpdate = false;
                        if (terminate)
                            break;
                    }

                    MPI_Request solutionRequest;
                    // Choose communication tag according to node status
                    COMMUNICATION_TAG nodeStatus = TAG_SOLVED_NODE_STATUS_NORMAL;
                    if (nodeConverged)
                        nodeStatus = TAG_SOLVED_NODE_STATUS_CONVERGED;
                    if (nodeProvenInfeasible)
                        nodeStatus = TAG_SOLVED_NODE_STATUS_INFEAS;

                    // Send communication request to manager
                    MPI_Issend(NULL, 0, MPI_INT, 0, nodeStatus, MPI_COMM_WORLD, &solutionRequest);
                    _sync_with_master(solutionRequest, terminate);
                    if (terminate) {
                        break;
                    }
                    _send_solved_problem(currentNode, currentLBD, lbpSolutionPoint, nAddedLBSolves, nAddedUBSolves, nodeStatus);

                MAiNGO_END_IF    // End of MAiNGO_IF_BAB_WORKER
#endif

#ifdef HAVE_MAiNGO_MPI
                MAiNGO_IF_BAB_MANAGER
                    // -----------------------------------
                    // Receive solved node from Worker
                    // -----------------------------------

                    MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    _workCount--;
                    _nodesGivenToWorkers[status.MPI_SOURCE - 1].first  = false;
                    _nodesGivenToWorkers[status.MPI_SOURCE - 1].second = _maingoSettings->infinity;

                    // Receive problem
                    COMMUNICATION_TAG nodeStatus = (COMMUNICATION_TAG)status.MPI_TAG;
                    std::vector<double> lbpSolutionPoint;
                    unsigned nAddedLBSolves;
                    unsigned nAddedUBSolves;
                    bool nodeConverged        = false;
                    bool nodeProvenInfeasible = false;
                    // Get node status
                    switch (nodeStatus) {
                        case TAG_SOLVED_NODE_STATUS_CONVERGED:
                            nodeConverged = true;
                            break;
                        case TAG_SOLVED_NODE_STATUS_INFEAS:
                            nodeProvenInfeasible = true;
                            break;
                        default:
                            break;
                    }

                    _recv_solved_problem(currentNode, currentLBD, lbpSolutionPoint, nAddedLBSolves, nAddedUBSolves, nodeStatus, status.MPI_SOURCE);
#endif

                    //-------------------------------------
                    // 6. Process the incumbent status of the processed node
                    //-------------------------------------
#ifndef HAVE_MAiNGO_MPI    // In the parallel version the worker already told the manager about the incumbent
                    if (foundNewFeasiblePoint) {
                        // Check if this feasible point is better than the incumbent & fathom by value dominance (first remaining tree and then current node):
                        _update_incumbent_and_fathom(ubpObjectiveValue, ubpSolutionPoint, currentNode.get_ID());
                    }
#endif
                    // Update incumbent node information
                    if (currentNode.get_ID() == _incumbentNodeId) {
                        currentNode.set_holds_incumbent(true);
                    }
                    else {
                        currentNode.set_holds_incumbent(false);
                    }


                    // -----------------------------------
                    // 7. Branching
                    // -----------------------------------
                    // Inform brancher about changes in node
                    _brancher->register_node_change(currentNode.get_ID(), currentNode);
                    // If node is feasible but it did not converge yet branch it
                    bool nodeReachedMinRelNodeSize = false;    // nodeReachedMinRelNodeSize indicates whether the brancher chose not to branch on the node because it is too small
                    if ((nodeProvenInfeasible == false) && (nodeConverged == false)) {
                        currentNode.set_pruning_score(currentLBD);
                        bool dummy;
                        std::tie(dummy, nodeReachedMinRelNodeSize) = _brancher->branch_on_node(currentNode, lbpSolutionPoint, currentLBD, _incumbentNodeId, _maingoSettings->relNodeTol);
                    }

                    // -----------------------------------
                    // 8. Heuristic checks
                    // -----------------------------------
                    if (!_moreScalingActivated) {
                        _check_if_more_scaling_needed();
                    }

                    // -----------------------------------
                    // 9. Updating statistics
                    // -----------------------------------
                    // Update best fathomed lower bound properly. This ensures that the final overall lower bound is correct.
                    // In particular, nodeReachedMinRelNodeSize is needed to ensure the overall LBD does not get better than that of nodes which were thrown out because they were too small.
                    if (nodeConverged || nodeReachedMinRelNodeSize) {
                        _bestLbdFathomed = std::min(_bestLbdFathomed, currentLBD);
                    }
                    // Update overall lbd
                    _update_lowest_lbd();
                    // Update counters
                    _iterations++;
                    _ubdcnt += nAddedUBSolves;
                    _lbdcnt += nAddedLBSolves;
                    if (nodeConverged || nodeProvenInfeasible) {
                        _nNodesFathomed++;
                    }
                    _nNodesLeft        = _brancher->get_nodes_in_tree();
                    _nNodesMaxInMemory = std::max(_nNodesMaxInMemory, _nNodesLeft);
                MAiNGO_END_IF    // End of MAiNGO_IF_BAB_MANAGER in line 325

                // -----------------------------------
                // 10. Timing & output
                // -----------------------------------
                // Get time
                currentTime = get_cpu_time();
                if ((currentTime - startTime) < lastTime) {
                    _daysPassed += 1;
                    MAiNGO_IF_BAB_MANAGER
                        std::ostringstream outstr;
                        outstr << "    Days spent: " << _daysPassed << std::endl
                               << std::endl;
                        _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
                    MAiNGO_END_IF
                }
                _timePassed = _timePreprocess + _daysPassed * 86400 + (currentTime - startTime);
                lastTime    = currentTime - startTime;
                MAiNGO_IF_BAB_MANAGER
                    // Print output
                    _display_and_log_progress(currentLBD, currentNode);

                    // -----------------------------------
                    // 11. Check termination
                    // -----------------------------------
                    termination = _check_termination();
                MAiNGO_END_IF

            }    // main while loop

            // ----------------- End 1-10: Main branch and bound loop ----------------------------

#ifdef HAVE_MAiNGO_MPI
            MAiNGO_IF_BAB_MANAGER
                // ---------------------------------------------------------------------------------
                // Inform workers about termination
                // ---------------------------------------------------------------------------------

                _inform_worker_about_event(BCAST_TERMINATE, /* blocking broadcast */ true);
                // Wait for all workers to get out of the while(true) loop
                MAiNGO_MPI_BARRIER
                    // Clean up all pending requests -- it is not the best way, but MPI_Cancel seems to have problems on different operating systems so we do this just to make sure everything is fine
                    bool pendingRequests = true;
                while (pendingRequests) {
                    int flag = 0;
                    MPI_Status status;
                    MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
                    if (flag) {
                        MPI_Recv(NULL, 0, MPI_INT, status.MPI_SOURCE, status.MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                    }
                    else {
                        pendingRequests = false;
                    }
                }
                MAiNGO_MPI_BARRIER

                    // Get CPU times of workers
                    std::vector<double>
                        processTimes(_nProcs);
                MPI_Gather(&_timePassed, 1, MPI_DOUBLE, processTimes.data(), 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

                // Calculate sum of compute times
                _timePassed = 0;
                for (int i = 0; i < _nProcs; i++) {
                    _timePassed += processTimes[i];
                }
                MAiNGO_ELSE
                    // ---------------------------------------------------------------------------------
                    // Inform manager about required solution time
                    // ---------------------------------------------------------------------------------
                    MAiNGO_MPI_BARRIER        // Wait for all workers to get out of the while(true) loop
                        MAiNGO_MPI_BARRIER    // Wait for manager to clean up all pending requests
                            // Send needed CPU time to master
                            MPI_Gather(&_timePassed, 1, MPI_DOUBLE, NULL, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
                MAiNGO_END_IF
#endif

                // ---------------------------------------------------------------------------------
                // 11: Prepare output
                // ---------------------------------------------------------------------------------
                MAiNGO_IF_BAB_MANAGER
                    // Store solution for returning to MAiNGO
                    if (_foundFeas == true) {
                        solutionPoint = _incumbent;
                        solutionValue = _ubd;
                    }

                    // Save time passed (this is especially needed when the passed CPU time is >24h)
                    timePassed = _timePassed;
                MAiNGO_END_IF
        // --------------------- End: Prepare output -------------------------------
    }
#ifdef HAVE_MAiNGO_MPI
    catch (std::exception &e) {
        MAiNGOMpiException e_mpi("  Error during branch-and-bound.", e, MAiNGOMpiException::ORIGIN_ME);
        _communicate_exception_and_throw(e_mpi);
    }
    catch (...) {
        std::string str;
        MAiNGO_IF_BAB_MANAGER
            str = "  Unknown error during branch-and-bound (Manager).";
            MAiNGO_ELSE
                str = "  Unknown error during branch-and-bound (Worker).";
            MAiNGO_END_IF
            MAiNGOMpiException e(str, MAiNGOMpiException::ORIGIN_ME);
            _communicate_exception_and_throw(e);
    }
#else
catch (std::exception &e)
{
    throw MAiNGOException("  Error during branch-and-bound.", e);
}
catch (...)
{
    throw MAiNGOException("  Unknown error during branch-and-bound.");
}
#endif
    return _status;    // Return code is determined in _check_termination()
}


////////////////////////////////////////////////////////////////////////////////////////
// function processing the current node (pre-processing, LBP, UBP, post-processing)
std::tuple<bool, bool, int, int, double, std::vector<double>, bool, double, std::vector<double>>
BranchAndBound::_process_node(babBase::BabNode &currentNode)
{


    // Prepare output variables
    bool nodeProvenInfeasible = false;
    bool nodeConverged        = false;
    int lbdcnt                = 0;    // output outside needs to increase _lbdcnt by lbdcnt
    int ubdcnt                = 0;    // see above
    double currentLBD         = currentNode.get_pruning_score();
    std::vector<double> lbpSolutionPoint;
    bool foundNewFeasiblePoint = false;
    double ubpObjectiveValue   = _maingoSettings->infinity;
    std::vector<double> ubpSolutionPoint;


    // -----------------------------------
    // 2. Node pre-processing: Constraint propagation and Optimization-based bound tightening (OBBT)
    // -----------------------------------
    nodeProvenInfeasible = _preprocess_node(currentNode);    // Tightens bounds and possibly proves infeasibilty

    if (nodeProvenInfeasible == false) {    // Only continue with 3-5 if node has not been proven to be infeasible

        // -----------------------------------
        // 3. Lower bounding: solve lower bounding problem (LBP) to derive a better lower bound
        // -----------------------------------
        lbp::LbpDualInfo dualInfo;
        std::tie(nodeProvenInfeasible, nodeConverged, currentLBD, lbpSolutionPoint, dualInfo) = _solve_LBP(currentNode);
        lbdcnt++;

        if ((nodeProvenInfeasible == false) && (nodeConverged == false)) {

            // -----------------------------------
            // 4. Upper bounding: try to find a good feasible point within the current node
            // -----------------------------------
            ubpSolutionPoint = lbpSolutionPoint;
            if (_maingoSettings->LBP_addAuxiliaryVars && !lbpSolutionPoint.empty()) {
                ubpSolutionPoint.resize(_nvarWOaux);
            }
            std::tie(foundNewFeasiblePoint, nodeConverged, ubpObjectiveValue) = _solve_UBP(currentNode, ubpSolutionPoint, currentLBD);
            ubdcnt++;
            if (nodeConverged == false) {

                // -----------------------------------
                // 5. Node post-processing: Duality-based bounds tightening (DBBT) & probing
                // -----------------------------------
                nodeProvenInfeasible = _postprocess_node(currentNode, lbpSolutionPoint, dualInfo);

            }    // end if(nodeConverged==false)

        }    // end if(((nodeProvenInfeasible==false)&&(nodeConverged==false))


    }    // end if(nodeProvenInfeasible==false)

    // Infeasible nodes get infinity as currentLBD
    if (nodeProvenInfeasible) {
        currentLBD = _maingoSettings->infinity;
    }

    return std::make_tuple(nodeProvenInfeasible, nodeConverged, lbdcnt, ubdcnt, currentLBD, lbpSolutionPoint, foundNewFeasiblePoint, ubpObjectiveValue, ubpSolutionPoint);
}


////////////////////////////////////////////////////////////////////////////////////////
// function for pre-processing the current node (constraint propagation + OBBT)
bool
BranchAndBound::_preprocess_node(babBase::BabNode &currentNode)
{

    bool nodeProvenInfeasible = false;

    // -----------------------------------
    // 2a Constraint propagation
    // -----------------------------------
    TIGHTENING_RETCODE conPropStatus = TIGHTENING_UNCHANGED;
    if (_maingoSettings->BAB_constraintPropagation) {
        if (_foundFeas == true) {    // If we already have a feasible point, consider an upper bound for the objective as well
            conPropStatus = _LBS->do_constraint_propagation(currentNode, _ubd);
        }
        else {    // If we do not have a feasible point, use infinity as upper bound
            conPropStatus = _LBS->do_constraint_propagation(currentNode, _maingoSettings->infinity);
        }
    }
    if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
        _print_one_node(currentNode.get_pruning_score(), currentNode);
    }

    // Check if constraint propagation proved node to be infeasible
    if (conPropStatus == TIGHTENING_INFEASIBLE) {

        nodeProvenInfeasible = true;
    }
    else {    // If constraint propagation did not prove node to be infeasible, proceed with obbt

        // -----------------------------------
        // 2b Optimization-based bound tightening (OBBT)
        // -----------------------------------
        TIGHTENING_RETCODE obbtStatus = TIGHTENING_UNCHANGED;
        if (_maingoSettings->BAB_alwaysSolveObbt) {
            if (_foundFeas == true) {    // If we already have a feasible point, consider optimality as well
                obbtStatus = _LBS->solve_OBBT(currentNode, _ubd, lbp::OBBT_FEASOPT);
            }
            else {    // If we do not have a feasible point, use only feasibility
                obbtStatus = _LBS->solve_OBBT(currentNode, _ubd, lbp::OBBT_FEAS);
            }
        }
        // Check if OBBT proved node to be infeasible
        if (obbtStatus == TIGHTENING_INFEASIBLE) {
            nodeProvenInfeasible = true;
        }
        if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
            _print_one_node(currentNode.get_pruning_score(), currentNode);
        }
    }

    return nodeProvenInfeasible;
}


////////////////////////////////////////////////////////////////////////////////////////
// function invoking the LBS to solve the lower bounding problem
std::tuple<bool /*nodeProvenInfeasible*/, bool /*nodeConverged*/, double /*currentLBD*/, std::vector<double> /*lbpSolutionPoint*/, lbp::LbpDualInfo /*dualInfo*/>
BranchAndBound::_solve_LBP(const babBase::BabNode &currentNode)
{

    bool nodeProvenInfeasible = false;
    bool nodeConverged        = false;
    const double parentLBD    = currentNode.get_pruning_score();
    double currentLBD         = parentLBD;
    std::vector<double> lbpSolutionPoint;
    lbp::LbpDualInfo dualInfo;

    SUBSOLVER_RETCODE lbpStatus = _LBS->solve_LBP(currentNode, currentLBD, lbpSolutionPoint, dualInfo);

    if (currentLBD < parentLBD) {
        if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
            std::ostringstream outstr;
            outstr << "  LBD obtained for node " << currentNode.get_ID() << " is lower than LBD of its parent node. Using parent LBD." << std::endl;
            _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_ALL, _maingoSettings->loggingDestination);
        }
        currentLBD = parentLBD;
    }

    // Check if LBP proved node to be infeasible
    if (lbpStatus == SUBSOLVER_INFEASIBLE) {    // Ok, exit DBBT loop and just forget node
        currentLBD           = _maingoSettings->infinity;
        nodeProvenInfeasible = true;
    }

    // Check if LBP proved that node cannot contain a better point
    if (babBase::larger_or_equal_within_rel_and_abs_tolerance(currentLBD, _ubd, _maingoSettings->epsilonR, _maingoSettings->epsilonA)) {    // Ok, fathomed by value dominance
        if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
            std::ostringstream outstr;
            outstr << "  Node #" << currentNode.get_ID() << " converged with LBD " << currentLBD << " to UBD " << _ubd << std::endl;
            _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_ALL, _maingoSettings->loggingDestination);
        }

        nodeConverged = true;
    }

    return std::make_tuple(nodeProvenInfeasible, nodeConverged, currentLBD, lbpSolutionPoint, dualInfo);
}


////////////////////////////////////////////////////////////////////////////////////////
// function invoking the UBS to solve the upper bounding problem
std::tuple<bool /*foundNewFeasiblePoint*/, bool /*nodeConverged*/, double /*ubpObjectiveValue*/>
BranchAndBound::_solve_UBP(const babBase::BabNode &currentNode, std::vector<double> &ubpSolutionPoint, const double currentLBD)
{

    bool foundNewFeasiblePoint  = false;
    bool nodeConverged          = false;
    double ubpObjectiveValue    = _maingoSettings->infinity;
    SUBSOLVER_RETCODE ubpStatus = _UBS->solve(currentNode, ubpObjectiveValue, ubpSolutionPoint);

    // Check if we found a feasible point
    if (ubpStatus == SUBSOLVER_FEASIBLE) {
        foundNewFeasiblePoint = true;

        if ((ubpObjectiveValue < currentLBD - _maingoSettings->epsilonA) && (ubpObjectiveValue < currentLBD - std::fabs(ubpObjectiveValue) * _maingoSettings->epsilonR)) {    // Sanity check to detect inconsistent bounding operations
            if (ubpObjectiveValue > -_maingoSettings->infinity) {
                std::ostringstream errmsg;
                errmsg << std::endl
                       << "  Error while checking objective returned by upper bounding solver: Upper bound < lower bound for the following node: " << std::endl;
                _print_one_node(currentNode.get_pruning_score(), currentNode, errmsg);
                errmsg << "  LBD = " << std::setprecision(16) << currentLBD << std::endl
                       << "UBD = " << ubpObjectiveValue << std::endl;
#ifdef HAVE_MAiNGO_MPI
                throw MAiNGOMpiException(errmsg.str(), currentNode, MAiNGOMpiException::ORIGIN_ME);
#else
                throw MAiNGOException(errmsg.str());
#endif
            }
            else {
                ubpObjectiveValue = _maingoSettings->infinity;
                std::ostringstream outstr;
                outstr << "  Warning: UBD found in node " << currentNode.get_ID() << " is lower than the MAiNGO infinity value " << -_maingoSettings->infinity << ".\n";
                outstr << "           Please consider scaling your objective function.\n";
                _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
            }
        }

#ifdef HAVE_MAiNGO_MPI
        // Check if this feasible point is better than the incumbent & inform manager about it
        if (ubpObjectiveValue < _ubd) {
            _send_incumbent(ubpObjectiveValue, ubpSolutionPoint, currentNode.get_ID());
        }
#endif

        // Check if this feasible point is better than the incumbent & fathom by value dominance (first remaining tree and then current node):
        if (babBase::larger_or_equal_within_rel_and_abs_tolerance(currentLBD, std::min(_ubd, ubpObjectiveValue), _maingoSettings->epsilonR, _maingoSettings->epsilonA)) {    // Ok, fathomed by value dominance
            if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
                std::ostringstream outstr;
                outstr << "  Node #" << currentNode.get_ID() << " converged with LBD " << currentLBD << " to UBD " << _ubd << std::endl;
                _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_ALL, _maingoSettings->loggingDestination);
            }
            nodeConverged = true;
        }
    }

    return std::make_tuple(foundNewFeasiblePoint, nodeConverged, ubpObjectiveValue);
}


////////////////////////////////////////////////////////////////////////////////////////
// Function for post-processing the current node. Includes bound DBBT and probing
bool
BranchAndBound::_postprocess_node(babBase::BabNode &currentNode, const std::vector<double> &lbpSolutionPoint, const lbp::LbpDualInfo &dualInfo)
{

    bool nodeProvenInfeasible = false;
    if ((dualInfo.multipliers.size() != _nvar) || (lbpSolutionPoint.size() != _nvar)) {    // We can only do DBBT or probing if we have multipliers and a solution point from LBP
        // No multipliers or no valid LBD available --> cannot proceed with post-processing --> node not converged, return
    }
    else {
        TIGHTENING_RETCODE status = _LBS->do_dbbt_and_probing(currentNode, lbpSolutionPoint, dualInfo, _ubd);
        if (_maingoSettings->BAB_verbosity > VERB_NORMAL) {
            _print_one_node(currentNode.get_pruning_score(), currentNode);
        }
        switch (status) {
            case TIGHTENING_INFEASIBLE:
                nodeProvenInfeasible = true;
                break;
            case TIGHTENING_UNCHANGED:
            case TIGHTENING_CHANGED:
                break;
        }
    }

    return nodeProvenInfeasible;
}


////////////////////////////////////////////////////////////////////////////////////////
// function updating the incumbent, sending the information to the LBS and fathoming nodes
void
BranchAndBound::_update_incumbent_and_fathom(const double solval, const std::vector<double> sol, const unsigned int currentNodeID)
{


    // Update incumbent value
    if (solval < _ubd) {
#ifdef HAVE_MAiNGO_MPI
        // Inform workers that a feasible point was found
        if (!_foundFeas) {
            _inform_worker_about_event(BCAST_FOUND_FEAS, /*non-blocking*/ false);
        }
#endif
        _foundFeas  = true;
        _firstFound = _iterations;
        // Only print line if new incumbent is better than the old one by more than the specified tolerance
        if (!babBase::larger_or_equal_within_rel_and_abs_tolerance(solval, _ubd, _maingoSettings->epsilonR, _maingoSettings->epsilonA)) {
            _printNewIncumbent = true;
        }

        _ubd             = solval;
        _incumbent       = sol;
        _incumbentNodeId = currentNodeID;
        _LBS->update_incumbent_LBP(_incumbent);

#ifdef HAVE_MAiNGO_MPI
        _informedWorkerAboutIncumbent = std::vector<bool>(_nProcs - 1, false);
#endif

        // Inform _brancher about new incumbent
        _brancher->set_new_incumbent_point(_incumbent);
        size_t nodesBefore         = _brancher->get_nodes_in_tree();
        double smallestFathomedLBD = _brancher->decrease_pruning_score_threshold_to(_ubd);    // Here, nodes with lbd exceeding the new _ubd are fathomed
        _bestLbdFathomed           = std::min(smallestFathomedLBD, _bestLbdFathomed);
        _nNodesFathomed += nodesBefore - _brancher->get_nodes_in_tree();
        _nNodesDeleted += nodesBefore - _brancher->get_nodes_in_tree();
        _nNodesLeft = _brancher->get_nodes_in_tree();
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function for updating the lowest lower bound
void
BranchAndBound::_update_lowest_lbd()
{

    _lbdOld = _lbd;
#ifdef HAVE_MAiNGO_MPI
    if (_brancher->get_nodes_in_tree() > 0 || _workCount > 0) {
#endif
        if (_brancher->get_nodes_in_tree() > 0) {
            // This is to make sure that in case we needed to throw a node out because of node size, the proven _lbd does not get better than this node
            _lbd = std::min(_brancher->get_lowest_pruning_score(), _bestLbdFathomed);
#ifdef HAVE_MAiNGO_MPI
            double minValWorkers = std::min_element(_nodesGivenToWorkers.begin(), _nodesGivenToWorkers.end(), maingo::WorkerNodeComparator())->second;
            _lbd                 = std::min(_lbd, minValWorkers);
#endif
        }
#ifdef HAVE_MAiNGO_MPI
        else {    // Case where there are no nodes in the tree but workers are still working
            _lbd                 = std::min(_lbd, _bestLbdFathomed);
            double minValWorkers = std::min_element(_nodesGivenToWorkers.begin(), _nodesGivenToWorkers.end(), maingo::WorkerNodeComparator())->second;
            _lbd                 = std::min(_lbd, minValWorkers);
        }
    }
#endif
    else {
        _lbd = _bestLbdFathomed;
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function which checks whether it is necessary to activate scaling within the LBD solver. This is a heuristic approach, which does not affect any deterministic optimization assumptions
void
BranchAndBound::_check_if_more_scaling_needed()
{

    if (!_logger->reachedMinNodeSize) {
        if (mc::isequal(_lbd, _lbdOld, mc::machprec(), mc::machprec())) {    // If the lbd did not change enough increase the counter
            _lbdNotChanged++;
        }
        else {
            _lbdNotChanged = 0;
        }
        // If the lbd did not change for at least LBP_activateMoreScaling iterations, scaling was not yet activated (checked before going into this function) and the difference between lowest lbd and current best ubd is small enough, we activate more scaling in the LBS
        if (_lbdNotChanged > _maingoSettings->LBP_activateMoreScaling && (_lbd >= (_ubd - 1E-2) || _lbd >= (_ubd - std::fabs(_ubd) * 1E-1))) {
            _LBS->activate_more_scaling();
            _moreScalingActivated = true;
            _lbdNotChanged        = 0;
#ifdef HAVE_MAiNGO_MPI
            _inform_worker_about_event(BCAST_SCALING_NEEDED, true);
#endif
            if (_maingoSettings->LBP_solver > 1) {    // LBP_solver = 0 is MAiNGO default, LBP_solver = 1 is interval-based
                VERB verbosityGiven = std::max(_maingoSettings->BAB_verbosity, _maingoSettings->LBP_verbosity);
                _logger->print_message("  Warning: Additional scaling in the lower bounding solver activated.\n", verbosityGiven, VERB_NORMAL, _maingoSettings->loggingDestination);
            }
        }
    }
}


////////////////////////////////////////////////////////////////////////////////////////
// function for displaying the rows of the B&B output and saving them to logs
void
BranchAndBound::_display_and_log_progress(const double currentNodeLBD, const babBase::BabNode &currentNode)
{

#ifndef HAVE_MAiNGO_MPI
    int _workCount = 0;
#endif
    // Determine whether to print due to a special occasion
    bool solved               = (_nNodesLeft == 0 && _workCount == 0);
    bool maxNodesReached      = (_nNodesLeft >= _maingoSettings->BAB_maxNodes && _workCount == 0);
    bool maxTimeReached       = (_timePassed >= _maingoSettings->maxTime && _workCount == 0);
    bool maxIterationsReached = (_iterations >= _maingoSettings->BAB_maxIterations && _workCount == 0);
    bool specialOccasion      = _printNewIncumbent || solved || maxNodesReached || maxTimeReached || maxIterationsReached || _iterations == 1;
    bool printToStream        = specialOccasion || (!((int)fmod((double)_iterations, (double)_maingoSettings->BAB_printFreq)));
    bool printToLog           = specialOccasion || (!((int)fmod((double)_iterations, (double)_maingoSettings->BAB_logFreq)));

    std::ostringstream strout;
    std::ostringstream stroutCsv;
    if (_maingoSettings->BAB_verbosity >= VERB_NORMAL) {
        if (_linesprinted == 0 || !((int)fmod((double)_linesprinted, (double)_iterationsgap))) {
            strout << "  " << std::setw(9) << "Iteration"
                   << "  "
#ifdef MAiNGO_DEBUG_MODE
                   << std::setw(9) << " NodeId  "
                   << "  "
                   << std::setw(15) << "NodeLBD  "
                   << "  "
#endif
                   << std::setw(15) << "LBD      "
                   << "  "
                   << std::setw(15) << "UBD      "
                   << "  "
                   << std::setw(9) << "NodesLeft"
                   << "  "
                   << std::setw(15) << "AbsGap   "
                   << "  "
                   << std::setw(15) << "RelGap   "
                   << "  "
                   << std::setw(15) << "CPU     "
                   << "  "
                   << std::endl;
            _linesprinted = 1;
        }
        if (printToStream || printToLog) {
            strout.setf(std::ios_base::scientific);
            stroutCsv.setf(std::ios_base::scientific);
            if (_printNewIncumbent && _firstFound != 0) {    // Mark line with * in case a new incumbent was found in this iteration
                strout << "  *";
            }
            else {
                strout << "   ";
            }
            strout << std::setw(8) << _iterations << "  "
#ifdef MAiNGO_DEBUG_MODE
                   << std::setw(9) << currentNode.get_ID() << "  "
                   << std::setw(15) << currentNodeLBD << "  "
#endif
                   << std::setw(15) << _lbd << "  "
                   << std::setw(15) << _ubd << "  "
                   << std::setw(9) << _nNodesLeft << "  "
                   << std::setw(15) << _ubd - _lbd << "  ";
            if (mc::isequal(_ubd, 0, mc::machprec(), mc::machprec())) {
                strout << std::setw(15) << "N/A"
                       << "  ";
            }
            else {
                strout << std::setw(15) << (_ubd - _lbd) / std::fabs(_ubd) << "  ";
            }
            strout << std::setw(15) << _timePassed << "  " << std::endl;
            strout.unsetf(std::ios_base::scientific);
            if (_maingoSettings->writeCsv) {
                stroutCsv << std::setw(8) << _iterations << ","
#ifdef MAiNGO_DEBUG_MODE
                          << std::setw(9) << currentNode.get_ID() << ","
                          << std::setw(15) << currentNodeLBD << ","
#endif
                          << std::setw(15) << _lbd << ","
                          << std::setw(15) << _ubd << ","
                          << std::setw(9) << _nNodesLeft << ","
                          << std::setw(15) << _ubd - _lbd << ",";
                if (mc::isequal(_ubd, 0, mc::machprec(), mc::machprec())) {
                    stroutCsv << std::setw(15) << "N/A"
                              << ",";
                }
                else {
                    stroutCsv << std::setw(15) << (_ubd - _lbd) / std::fabs(_ubd) << ",";
                }
                stroutCsv << std::setw(15) << _timePassed << "," << std::endl;
                stroutCsv.unsetf(std::ios_base::scientific);
            }
        }
        _linesprinted++;
        if (printToStream) {
            // This is done instead of print_message, since the logging freq and print freq may differ
            _logger->print_message_to_stream_only(strout.str(), _maingoSettings->loggingDestination);
        }
    }

    // Save information if we want a log or csv file
    if (specialOccasion || printToLog) {
        if ((_maingoSettings->loggingDestination == LOGGING_FILE) || (_maingoSettings->loggingDestination == LOGGING_FILE_AND_STREAM)) {
            _logger->babLine.push(strout.str());
        }
        if (_maingoSettings->writeCsv) {
            _logger->babLineCsv.push(stroutCsv.str());
        }
    }
    if ((_maingoSettings->writeToLogSec > 0) && (((double)_timePassed) / ((double)_maingoSettings->writeToLogSec) > _writeToLogEverySec)) {
        if ((_maingoSettings->loggingDestination == LOGGING_FILE) || (_maingoSettings->loggingDestination == LOGGING_FILE_AND_STREAM)) {
            _logger->write_all_lines_to_log();
        }
        if (_maingoSettings->writeCsv) {
            _logger->write_all_iterations_to_csv();
        }
        _writeToLogEverySec++;
    }
    _printNewIncumbent = false;
}


////////////////////////////////////////////////////////////////////////////////////////
// function for printing exactly one node
void
BranchAndBound::_print_one_node(const double theLBD, const int ID, const std::vector<double> lowerVarBounds, const std::vector<double> upperVarBounds)
{
    std::ostringstream outstr;
    outstr << "  NODE " << ID << "  has lbd (inherited from parent) =" << std::setprecision(16) << theLBD << std::endl;
    for (unsigned int i = 0; i < _nvar; i++) {
        outstr << "  " << std::setprecision(16) << "var " << i + 1 << " " << lowerVarBounds[i] << "..." << upperVarBounds[i] << std::endl;
    }
    _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_ALL, _maingoSettings->loggingDestination);
}


////////////////////////////////////////////////////////////////////////////////////////
// function for printing exactly one node
void
BranchAndBound::_print_one_node(const double theLBD, const int ID, const std::vector<double> lowerVarBounds, const std::vector<double> upperVarBounds, std::ostream &outstream)
{
    std::ostringstream outstr;
    outstr << "  NODE " << ID << "  has lbd (inherited from parent) =" << std::setprecision(16) << theLBD << std::endl;
    for (unsigned int i = 0; i < _nvar; i++) {
        outstr << std::setprecision(16) << "var " << i + 1 << " " << lowerVarBounds[i] << "..." << upperVarBounds[i] << std::endl;
    }
    outstream << outstr.str();
}


////////////////////////////////////////////////////////////////////////////////////////
// function for checking termination
BranchAndBound::_TERMINATION_TYPE
BranchAndBound::_check_termination()
{

#ifdef HAVE_MAiNGO_MPI
    MAiNGO_IF_BAB_WORKER
        // Workers get their termination signal from the manager, which causes the bab-loop to break
        return _NOT_TERMINATED;
    MAiNGO_END_IF
#else
    unsigned _workCount = 0;    // Dummy for convenience
#endif

    _TERMINATION_TYPE terminate = _NOT_TERMINATED;
    bool solved                 = (_nNodesLeft == 0 && _workCount == 0);
    bool maxNodesReached        = (_nNodesLeft >= _maingoSettings->BAB_maxNodes);
    bool maxTimeReached         = (_timePassed >= _maingoSettings->maxTime);
    bool maxIterationsReached   = (_iterations >= _maingoSettings->BAB_maxIterations);
    bool reachedTargetUbd       = (_ubd <= _maingoSettings->targetUpperBound);
    bool reachedTargetLbd       = (_lbd >= _maingoSettings->targetLowerBound);
    unsigned *limit;
    std::string criterion;
    std::string message;
    std::string userInput;


    if (solved) {    // Regular termination.
        if (_foundFeas) {
            // If we terminated regularly, we either have a global optimum, we reached min node size or the problem is infeasible
            if (babBase::larger_or_equal_within_rel_and_abs_tolerance(_lbd, _ubd, _maingoSettings->epsilonR, _maingoSettings->epsilonA)) {
                _status = babBase::enums::GLOBALLY_OPTIMAL;
            }
            else {
                _status = babBase::enums::GLOBAL_MIN_NODE_SIZE;
            }
        }
        else {
            _status = babBase::enums::INFEASIBLE;
        }
        terminate = _TERMINATED;

        if (_status == babBase::enums::GLOBALLY_OPTIMAL || _status == babBase::enums::INFEASIBLE) {
            // Print on screen and write to log
            _logger->print_message("  Done.\n", _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
            _print_termination("*                                             *** Regular termination. ***                                             *");
        }
        else {
            // Print on screen and write to log
            _logger->print_message("  Done.\n", _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
            _print_termination("*                                             *** Regular termination. ***                                             *\n*               *** Reached minimum node size. User-defined optimality tolerance could not be reached ***              *");
        }
    }
    else if (reachedTargetUbd) {    // Reached user-defined upper bound
        if (_workCount > 0) {
            return _TERMINATED_WORKERS_ACTIVE;
        }
        terminate = _TERMINATED;
        _status   = babBase::enums::TARGET_UBD;
        _logger->print_message("  Done.\n", _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
        _print_termination("*                                          *** Reached target upper bound. ***                                         *");
    }
    else if (reachedTargetLbd) {    // Reached user-defined lower bound
        if (_workCount > 0) {
            return _TERMINATED_WORKERS_ACTIVE;
        }
        terminate = _TERMINATED;
        _status   = babBase::enums::TARGET_LBD;
        _logger->print_message("  Done.\n", _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
        _print_termination("*                                          *** Reached target lower bound. ***                                         *");
    }
    else if (_maingoSettings->terminateOnFeasiblePoint && _foundFeas) {    // Found feasible point and this is enough for the user
        if (_workCount > 0) {
            return _TERMINATED_WORKERS_ACTIVE;
        }
        terminate = _TERMINATED;
        _status   = babBase::enums::FEASIBLE_POINT_ONLY;
        _logger->print_message("  Done.\n", _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
        _print_termination("*                                            *** Found feasible point. ***                                             *");
    }
    else if (maxNodesReached || maxTimeReached || maxIterationsReached) {    // Reached some other termination criterion.

        if (_workCount > 0) {
            return _TERMINATED_WORKERS_ACTIVE;
        }
        babBase::enums::BAB_RETCODE potentialStatus;
        if (maxNodesReached) {
            message         = "*                                   *** Reached maximum number of nodes in memory! ***                                 *";
            criterion       = "number of nodes in memory";
            limit           = &_maingoSettings->BAB_maxNodes;
            potentialStatus = babBase::enums::MAX_NODES;
        }
        else if (maxTimeReached) {
            message         = "*                                              *** Reached CPU time limit! ***                                             *";
            criterion       = "CPU time";
            limit           = &_maingoSettings->maxTime;
            potentialStatus = babBase::enums::MAX_TIME;
        }
        else {
            message         = "*                                      *** Reached maximum number of B&B iterations! ***                                   *";
            criterion       = "B&B iterations";
            limit           = &_maingoSettings->BAB_maxIterations;
            potentialStatus = babBase::enums::MAX_ITERATIONS;
        }


        if (!_maingoSettings->confirmTermination) {    // Just terminate.

            terminate = _TERMINATED;
            _print_termination(message);
            _status = potentialStatus;
        }
        else {    // Ask user whether to terminate

            _print_termination(message);

            // Summarize current status
            std::ostringstream outstr;
            outstr << std::endl
                   << "  " << std::setw(16) << std::left << " " << std::setw(10) << std::right << "Current" << std::setw(3) << " " << std::setw(10) << std::left << "Limit" << std::endl
                   << "  " << std::setw(16) << std::left << "Nodes in memory:" << std::setw(10) << std::right << _nNodesLeft << std::setw(3) << " " << std::setw(10) << std::left << _maingoSettings->BAB_maxNodes << std::endl
                   << "  " << std::setw(16) << std::left << "Iterations:" << std::setw(10) << std::right << _iterations << std::setw(3) << " " << std::setw(10) << std::left << _maingoSettings->BAB_maxIterations << std::endl
                   << "  " << std::setw(16) << std::left << "CPU time [s]:" << std::setw(10) << std::right << _timePassed << std::setw(3) << " " << std::setw(10) << std::left << _maingoSettings->maxTime << std::endl
                   << std::endl;
            outstr << "  " << std::setw(16) << std::left << "LBD:" << std::setw(10) << std::right << _lbd << std::endl
                   << "  " << std::setw(16) << std::left << "UBD:" << std::setw(10) << std::right << _ubd << std::endl
                   << "  " << std::setw(16) << std::left << "Absolute gap:" << std::setw(10) << std::right << _ubd - _lbd << std::setw(3) << " " << std::setw(10) << std::left << _maingoSettings->epsilonA << std::endl
                   << "  " << std::setw(16) << std::left << "Relative gap:" << std::setw(10) << std::right << ((_ubd == 0) ? (_lbd) : ((_ubd - _lbd) / std::fabs(_ubd))) << std::setw(3) << " " << std::setw(10) << std::left << _maingoSettings->epsilonR << std::endl
                   << std::endl
                   << "************************************************************************************************************************" << std::endl
                   << std::endl;

            // Query input
            outstr << "  Do you want to continue (y/n)? ";
            _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
            while (getline(std::cin, userInput) && userInput != "y" && userInput != "n") {
                _logger->print_message("  Invalid input. Please type 'y' or 'n' and press enter: ", _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
            }

            // Interpret input
            if (userInput == "y") {
                outstr.str("");
                outstr.clear();
                outstr << "  User input: yes\n  Enter new limit for " << criterion << ": ";
                _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
                while (getline(std::cin, userInput) && (*limit >= atof(userInput.c_str()))) {
                    outstr.str("");
                    outstr.clear();
                    outstr << "  Invalid input (has to be greater than  " << *limit << "), please revise: ";
                    _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
                }
                if (userInput == "") {
                    throw MAiNGOException("  Error while checking termination: Querying user input while checking termination criteria.");
                }
                else if (atof(userInput.c_str()) > std::numeric_limits<int>::max()) {
                    _logger->print_message("  Value is too high for integer type, set to maximum.\n", _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
                    *limit = std::numeric_limits<int>::max();
                }
                else {
                    *limit = (int)atof(userInput.c_str());
                }
                outstr.str("");
                outstr.clear();
                outstr << "  User input: " << (*limit) << "\n  Okay, Resuming solution..." << std::endl
                       << std::endl;
                _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
                _linesprinted = 0;    // To make sure header line is printed before continuing to print progress
                terminate     = _NOT_TERMINATED;
            }
            else if (userInput == "n") {
                _logger->print_message("  User input: no\n  ", _maingoSettings->BAB_verbosity, VERB_NONE, _maingoSettings->loggingDestination);
                terminate = _TERMINATED;
                _print_termination(message);
                _status = potentialStatus;
            }
            else {
                throw MAiNGOException("  Error while checking termination: invalid user input '" + userInput + "'");
            }
        }
    }
    return terminate;
}


////////////////////////////////////////////////////////////////////////////////////////
// function for printing stars in the termination
void
BranchAndBound::_print_termination(std::string message)
{

    std::ostringstream outstr;
    if ((_maingoSettings->BAB_verbosity > VERB_NONE) || (_maingoSettings->confirmTermination)) {
        outstr << std::endl
               << "************************************************************************************************************************" << std::endl
               << "*                                                                                                                      *" << std::endl
               << message << std::endl
               << "*                                                                                                                      *" << std::endl
               << "************************************************************************************************************************" << std::endl;
    }

    _logger->print_message(outstr.str(), _maingoSettings->BAB_verbosity, VERB_NORMAL, _maingoSettings->loggingDestination);
}